/******************************************************************************
* MagDock 1.2 - A Screen Magnifier Dock
* Copyright © JakeSFR 2019
* GNU GPL v3 applies (http://www.gnu.org/licenses/gpl.html)
******************************************************************************/

#include <X11/extensions/Xfixes.h>
#include <gtk/gtk.h>
#include <zlib.h>

// Gettext
#include <libintl.h>
#include <locale.h>
#define _(STRING) gettext(STRING)

/* ========================================================================= */

// Defaults

#define ZOOM			2.0			// zoom level (see ZOOM_LEVELS below)
#define POSITION		TOP			// position on the screen (TOP/BOTTOM)
#define OFFSET			0			// top/bottom offset (px)
#define HEIGHT			20			// magnifier height (% of screen height)
#define FPS				25			// frames per second
#define BGCOLOR			0x000000	// Background color (RGB) [not used in GUI]
#define SHOW_MAGNIFIER	TRUE		// show the magnifier or not
#define HIDE_ON_HOVER	TRUE		// hide magnifier on mouse hover or not
#define SHOW_CURSOR		TRUE		// show mouse cursor or not
#define INTERPOLATION	FALSE		// enable interpolation or not

/* ========================================================================= */

#define ZOOM_LEVELS		1.5, 2.0, 3.0, 4.0, 5.0, 7.5, 10.0
#define OFFSET_MIN		0
#define OFFSET_MAX		128
#define OFFSET_STEP		1
#define HEIGHT_MIN		10
#define HEIGHT_MAX		50
#define HEIGHT_STEP		5
#define FPS_MIN			10
#define FPS_MAX			50
#define FPS_STEP		5

/* ------------------------------------------------------------------------- */

#define APPNAME		"MagDock-1.2"
#define CONF_FILE	"magdock.conf"
#define EMBED_ICON	TRUE

#define TOP		0
#define BOTTOM	1

#define IF_OUTOF_RANGE(var, min, max, def) if( ((var) < (min)) || ((var) > (max)) ) (var) = (def)

// Globals
GtkWidget	*mag_window;
gboolean	mag_restart = FALSE;
gdouble		zoom_levels[] = { ZOOM_LEVELS };
size_t		zoom_levels_size;
guint		zoom_levels_current = 0;

struct {

	struct {
		gint width;
		gint height;
	} screen;

	struct {
		gint width;
		gint height;
	} win;
	
	struct {
		gint width;
		gint height;
	} mag;

	struct {
		GdkPixbuf	*pixbuf;
		gint		width;
		gint		height;
	} area;

} dim;

struct {
	gint		position;
	gint		offset;
	gint		height;
	gdouble		zoom;
	gint		fps;
	gint		bgcolor;
	gboolean	show_magnifier;
	gboolean	hide_on_hover;
	gboolean	show_cursor;
	gboolean	interpolation;
} option;

/* ========================================================================= */

void show_magnifier (void);

/* ------------------------------------------------------------------------- */

void check_error (GError **error) {

	if (*error) g_printerr ("Error: %s\n\n", (*error)->message);
	g_clear_error (error);

}

/* ------------------------------------------------------------------------- */

void load_defaults (void) {

	option.zoom					= ZOOM;
	option.position				= POSITION;
	option.offset				= OFFSET;
	option.height				= HEIGHT;
	option.fps					= FPS;
	option.bgcolor				= BGCOLOR;
	option.show_magnifier		= SHOW_MAGNIFIER;
	option.hide_on_hover		= HIDE_ON_HOVER;
	option.show_cursor			= SHOW_CURSOR;
	option.interpolation		= INTERPOLATION;

}

/* ------------------------------------------------------------------------- */

void load_options (void) {

	gchar		*filename;
	GKeyFile	*config;
	GError		*error = NULL;

	load_defaults();

	filename = g_build_filename (g_get_user_config_dir (), CONF_FILE, NULL);

	if (!g_file_test (filename, G_FILE_TEST_EXISTS)) {

		g_free (filename);

	} else {

		config = g_key_file_new ();
		g_key_file_load_from_file (config, filename, G_KEY_FILE_KEEP_COMMENTS, &error);
	
		if (error == NULL && g_key_file_has_group (config, "Settings")) {

			option.zoom				= g_key_file_get_double  (config, "Settings", "zoom", NULL);
			option.position			= g_key_file_get_integer (config, "Settings", "position", NULL);
			option.offset			= g_key_file_get_integer (config, "Settings", "offset", NULL);
			option.height			= g_key_file_get_integer (config, "Settings", "height", NULL);
			option.fps				= g_key_file_get_integer (config, "Settings", "fps", NULL);
			option.show_magnifier	= g_key_file_get_boolean (config, "Settings", "show_magnifier", NULL);
			option.hide_on_hover	= g_key_file_get_boolean (config, "Settings", "hide_on_hover", NULL);
			option.show_cursor		= g_key_file_get_boolean (config, "Settings", "show_cursor", NULL);
			option.interpolation	= g_key_file_get_boolean (config, "Settings", "interpolation", NULL);

		} else {

			check_error (&error);

		}
	
		g_key_file_free (config);
		g_free (filename);
	}

	// Sanity Checks
	zoom_levels_size = (sizeof zoom_levels) / (sizeof zoom_levels[0]);

	IF_OUTOF_RANGE (option.zoom, zoom_levels[0], zoom_levels[zoom_levels_size-1], ZOOM);
	IF_OUTOF_RANGE (option.position, 0, 1, POSITION);
	IF_OUTOF_RANGE (option.offset, OFFSET_MIN, OFFSET_MAX, OFFSET);
	IF_OUTOF_RANGE (option.height, HEIGHT_MIN, HEIGHT_MAX, HEIGHT);
	IF_OUTOF_RANGE (option.fps, FPS_MIN, FPS_MAX, FPS);
	IF_OUTOF_RANGE (option.show_magnifier, 0, 1, SHOW_MAGNIFIER);
	IF_OUTOF_RANGE (option.hide_on_hover, 0, 1, HIDE_ON_HOVER);
	IF_OUTOF_RANGE (option.show_cursor, 0, 1, SHOW_CURSOR);
	IF_OUTOF_RANGE (option.interpolation, 0, 1, INTERPOLATION);

	while((zoom_levels_current < zoom_levels_size) && (zoom_levels[zoom_levels_current] != option.zoom))
		zoom_levels_current++;

}

/* ------------------------------------------------------------------------- */

void save_options (void) {

	gchar		*filename;
	GKeyFile	*config;
	GError		*error = NULL;
	gsize		len;
	gchar		*contents;

	filename = g_build_filename (g_get_user_config_dir(), CONF_FILE, NULL);
	
	config = g_key_file_new ();

	g_key_file_set_double  (config,	"Settings", "zoom",				option.zoom);
	g_key_file_set_integer (config,	"Settings", "position",			option.position);
	g_key_file_set_integer (config,	"Settings", "offset",			option.offset);
	g_key_file_set_integer (config,	"Settings", "height",			option.height);
	g_key_file_set_integer (config,	"Settings", "fps",				option.fps);
	g_key_file_set_boolean (config, "Settings", "show_magnifier",	option.show_magnifier);
	g_key_file_set_boolean (config, "Settings", "hide_on_hover",	option.hide_on_hover);
	g_key_file_set_boolean (config, "Settings", "show_cursor",		option.show_cursor);
	g_key_file_set_boolean (config, "Settings", "interpolation",	option.interpolation);

	// Older glib doesn't have this useful function!
	//g_key_file_save_to_file (config, filename, &error);
	contents = g_key_file_to_data (config, &len, &error);
	check_error (&error);
	g_file_set_contents (filename, contents, len, &error);
	check_error (&error);

	g_key_file_free (config);
	g_free (filename);
	g_free (contents);

}

/* ------------------------------------------------------------------------- */

void create_area_pixbuf (void) {

	gboolean alpha;
	
	if (dim.area.pixbuf)
		g_object_unref(dim.area.pixbuf);

	dim.screen.width = gdk_screen_width ();
	dim.screen.height = gdk_screen_height ();

	dim.mag.width = dim.screen.width;
	dim.mag.height = (dim.screen.height / 100) * option.height;

	dim.area.width = dim.mag.width / option.zoom;
	dim.area.height = dim.mag.height / option.zoom;

	// If SHOW_CURSOR = FALSE, alpha can also be set to FALSE,
	// what makes rendering a little less CPU intensive.
	alpha = (option.show_cursor == TRUE) ? TRUE : FALSE;

	dim.area.pixbuf = gdk_pixbuf_new (
		GDK_COLORSPACE_RGB,
		alpha,
		8,
		dim.area.width,
		dim.area.height
	);

}
	
/* ------------------------------------------------------------------------- */

void move_magnifier (void) {

	dim.screen.width = gdk_screen_width ();
	dim.screen.height = gdk_screen_height ();

	gtk_window_get_size (GTK_WINDOW (mag_window), &dim.win.width, &dim.win.height);

	if (option.position == TOP) {

		gtk_window_move (GTK_WINDOW (mag_window), 0, 0 + option.offset);

	} else {

		gtk_window_move (GTK_WINDOW (mag_window), 0, dim.screen.height - dim.win.height - option.offset);

	}

}

/* ------------------------------------------------------------------------- */

void get_screen (GdkWindow *root_window, gint mouse_x, gint mouse_y, gint *offset_x, gint *offset_y) {

	gint trim_src_y, trim_dest_y, trim_height, win_y;

	// Calculate which portion of screen to grab
	if ( (mouse_x - dim.area.width / 2) < 0) {

		*offset_x = dim.area.width / 2;

	} else if ( (mouse_x + dim.area.width / 2) >= dim.screen.width) {

		*offset_x = dim.screen.width - dim.area.width / 2 - dim.area.width%2;	//%2 for off-by-one

	} else {

		*offset_x = mouse_x;

	}

	if ( (mouse_y - dim.area.height / 2) < 0) {

		*offset_y = dim.area.height / 2;

	} else if ( (mouse_y + dim.area.height / 2) >= dim.screen.height) {

		*offset_y = dim.screen.height - dim.area.height / 2 - dim.area.height%2;	//%2 for off-by-one

	} else {

		*offset_y = mouse_y;

	}

	// If mouse pointer is close to magnifier, trim the "feedback" area
	gtk_window_get_position (GTK_WINDOW (mag_window), NULL, &win_y);

	if ((mouse_y > win_y + dim.mag.height / 2) &&
		(mouse_y < dim.area.height / 2 + dim.win.height + win_y + (mouse_y - *offset_y))) {

		trim_src_y = trim_height = trim_dest_y = ABS(dim.area.height / 2 + dim.win.height + win_y - mouse_y + (mouse_y - *offset_y));

	} else if ((mouse_y <= win_y + dim.mag.height / 2) &&
		(mouse_y >= win_y - dim.area.height / 2 + (mouse_y - *offset_y))) {

		trim_src_y = trim_dest_y = 0;
		trim_height = ABS(win_y - mouse_y - dim.area.height / 2 + (mouse_y - *offset_y) - dim.area.height%2);

	} else {

		trim_src_y = trim_height = trim_dest_y = 0;

	}

	// Fill pixbuf with color. NB: add 0xff alpha.
	gdk_pixbuf_fill (dim.area.pixbuf, (option.bgcolor << 8) | 0xff);

	// Get pixels from screen
    gdk_pixbuf_get_from_drawable (
		dim.area.pixbuf,
		root_window,
		NULL,
		*offset_x - (dim.area.width / 2),
		*offset_y - (dim.area.height / 2) + trim_src_y,
		0, trim_dest_y,
		dim.area.width,
		dim.area.height - trim_height
	);

}

/* ------------------------------------------------------------------------- */

void get_cursor (gint mouse_x, gint mouse_y, gint offset_x, gint offset_y) {

	Display		*display;
	gint		event, error;
	guchar		*cur_conv;
	GdkPixbuf	*cur_pix;
	gint		trim_left, trim_right, trim_top, trim_bottom;

	display = XOpenDisplay (NULL);

	if (XFixesQueryExtension (display, &event, &error)) {

		XFixesCursorImage *cursor = XFixesGetCursorImage (display);

		// Convert cursor image to something usable in pixbuf
		cur_conv = g_malloc (cursor->width * cursor->height * 4);

		for (gint i = 0; i < (cursor->width * cursor->height); ++i) {
			cur_conv[0+i*4] = (guchar)((cursor->pixels[i] >> 16) & 0xff);
			cur_conv[1+i*4] = (guchar)((cursor->pixels[i] >>  8) & 0xff);
			cur_conv[2+i*4] = (guchar)((cursor->pixels[i] >>  0) & 0xff);
			cur_conv[3+i*4] = (guchar)((cursor->pixels[i] >> 24) & 0xff);
		}

		cur_pix = gdk_pixbuf_new_from_data (
			(guchar*)cur_conv,
			GDK_COLORSPACE_RGB, TRUE, 8,
			cursor->width, cursor->height,
			cursor->width * 4,
			NULL, NULL
		);

		// If cursor is larger than area, needs to be trimmed
		trim_left = trim_right = trim_top = trim_bottom = 0;

		// Trim Top
		if((cursor->height - (cursor->height - cursor->yhot)) > (mouse_y - offset_y + dim.area.height / 2)) {
			trim_top = (cursor->height - (cursor->height - cursor->yhot)) - (mouse_y - offset_y + dim.area.height / 2);
		}

		// Trim Bottom
		if((cursor->height - cursor->yhot) > (offset_y - mouse_y + dim.area.height / 2)) {
			trim_bottom = (cursor->height - cursor->yhot) - (offset_y - mouse_y + dim.area.height / 2) - dim.area.height%2;
		}

		// Trim Left
		if((cursor->width - (cursor->width - cursor->xhot)) > (mouse_x - offset_x + dim.area.width / 2)) {
			trim_left = (cursor->width - (cursor->width - cursor->xhot)) - (mouse_x - offset_x + dim.area.width / 2);
		}

		// Trim Right
		if((cursor->width - cursor->xhot) > (offset_x - mouse_x + dim.area.width / 2)) {
			trim_right = (cursor->width - cursor->xhot) - (offset_x - mouse_x + dim.area.width / 2) - dim.area.width%2;
		}

		// Add cursor image
		gdk_pixbuf_composite (
			cur_pix,
			dim.area.pixbuf,
			(dim.area.width / 2 - cursor->xhot) + mouse_x - offset_x + trim_left,	// dest_x
			(dim.area.height / 2 - cursor->yhot) + mouse_y - offset_y + trim_top,	// dest_y
			cursor->width - trim_left - trim_right,									// dest_width
			cursor->height - trim_top - trim_bottom,								// dest_height
			(dim.area.width / 2 - cursor->xhot) + mouse_x - offset_x,				// offset_x
			(dim.area.height / 2 - cursor->yhot) + mouse_y - offset_y,				// offset_y
			1, 1,
			GDK_INTERP_NEAREST,
			255
		);

		g_free (cursor);
		g_free (cur_conv);
		g_object_unref (cur_pix);

	}
	
	XCloseDisplay (display);

}

/* ------------------------------------------------------------------------- */

gint on_timeout (gpointer image) {

    GdkWindow	*root_window;
    gint		screen_width, screen_height;
    gint		mouse_x, mouse_y;
    gint		offset_x, offset_y;

	static gulong old_checksum = 0;
	gulong new_checksum;
	GdkPixbuf *area_final;

	// Restart magnifier on changing height, fps, cursor, etc.
	if (mag_restart == TRUE) {
		gtk_widget_destroy (mag_window);
		if (option.show_magnifier == TRUE) show_magnifier ();
		old_checksum = -1;
		mag_restart = FALSE;
		return FALSE;
	}

	// Magnifier window disabled, so stop the timeout
	if (option.show_magnifier == FALSE) {
		old_checksum = -1;	// reset checksum, so next time mag gets enabled, it also gets refreshed
		return FALSE;
	}

	// Restart magnifier window if screen resolution has changed
	screen_width = gdk_screen_width ();
	screen_height = gdk_screen_height ();

	if ((screen_width == dim.screen.width) && (screen_height == dim.screen.height)) {
		dim.screen.width = screen_width;
		dim.screen.height = screen_height;
	} else {
		mag_restart = TRUE;
		return TRUE;
	}

	// If mouse pointer enters the area of magnifier, hide it
    root_window = gdk_get_default_root_window ();
	gdk_window_get_pointer (root_window, &mouse_x, &mouse_y, NULL);

	gtk_window_get_size (GTK_WINDOW (mag_window), &dim.win.width, &dim.win.height);

	dim.mag.width = dim.screen.width;
	dim.mag.height = (dim.screen.height / 100 ) * option.height;

	if (option.hide_on_hover == TRUE) {

		if (((option.position == TOP ) &&
				(mouse_y < dim.win.height + option.offset) &&
				(mouse_y >= option.offset)) ||
			((option.position == BOTTOM) &&
				(mouse_y >= dim.screen.height - dim.win.height - option.offset) &&
				(mouse_y < dim.screen.height - option.offset))) {

			gtk_widget_hide (GTK_WIDGET (mag_window));

		} else {

			move_magnifier();
			gtk_widget_show (GTK_WIDGET (mag_window));

		}

	}

	// If magnifier window is hidden, leave here
	if (!gtk_widget_get_visible (mag_window)) return TRUE;

	// Get screen area
	get_screen (root_window, mouse_x, mouse_y, &offset_x, &offset_y);

	// Composite cursor image
	if (option.show_cursor) {

		get_cursor (mouse_x, mouse_y, offset_x, offset_y);

	}

	// Drawing to window every frame is very expensive (X process spikes),
	// so draw only if something has changed. Zlib's adler32 is quite fast.
	new_checksum = adler32 (
		adler32 (0L, Z_NULL, 0),
		gdk_pixbuf_get_pixels(dim.area.pixbuf),
		dim.area.width * dim.area.height * 4
	);

	if (old_checksum != new_checksum) {

		old_checksum = new_checksum;

		// Scale (GDK_INTERP_NEAREST, GDK_INTERP_TILES, GDK_INTERP_BILINEAR, GDK_INTERP_HYPER)
		area_final = gdk_pixbuf_scale_simple (
			dim.area.pixbuf,
			dim.mag.width,
			dim.mag.height,
			(option.interpolation != TRUE) ? GDK_INTERP_NEAREST : GDK_INTERP_BILINEAR
		);

		// Draw to window
		gtk_image_set_from_pixbuf (image, area_final);
		g_object_unref (area_final);

	}

	// Keep timeout going
	return TRUE;

}

/* ------------------------------------------------------------------------- */

void cb_disable_magnifier (GtkWidget *window, GdkEventButton *event, gpointer *data) {

	if (event->button == 2) {

		gtk_widget_destroy (mag_window);
		option.show_magnifier = FALSE;

	}

}

/* ------------------------------------------------------------------------- */

gboolean cb_delete_event (GtkWidget *window, GdkEventButton *event, gpointer *data) {

	// Special case for JWM (and apparently also i3) window manager,
	// which ignores the accept_focus property,
	// what allows closing the magnifier window by pressing Alt+F4

	return TRUE;

}

/* ------------------------------------------------------------------------- */

void show_magnifier (void) {

//	GtkWidget *mag_window;	// It's global!
	GtkWidget *separator_top, *separator_bottom, *vbox, *image;
	gulong vals[12] = { 0 };

	dim.screen.width = gdk_screen_width();
	dim.screen.height = gdk_screen_height();

	/* Window */
	mag_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (mag_window), APPNAME);
	gtk_window_set_keep_above (GTK_WINDOW (mag_window), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (mag_window), FALSE);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW (mag_window), TRUE);
	gtk_window_set_skip_pager_hint (GTK_WINDOW (mag_window), TRUE);
	gtk_window_set_accept_focus (GTK_WINDOW (mag_window), FALSE);
	gtk_window_set_focus_on_map (GTK_WINDOW (mag_window), FALSE);
	gtk_window_stick (GTK_WINDOW (mag_window));
	gtk_window_set_type_hint (GTK_WINDOW (mag_window), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_widget_set_events (mag_window, gtk_widget_get_events (mag_window) | GDK_BUTTON_PRESS_MASK);
	g_signal_connect (G_OBJECT (mag_window), "button-press-event", G_CALLBACK (cb_disable_magnifier), NULL);
	g_signal_connect (G_OBJECT (mag_window), "delete-event", G_CALLBACK (cb_delete_event), NULL);

	/* Vbox */
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (mag_window), vbox);

	/* Top Separator */
	separator_top = gtk_hseparator_new ();
	gtk_box_pack_start (GTK_BOX (vbox), separator_top, FALSE, FALSE, 0);

	/* Image */
	image = gtk_image_new ();
	gtk_box_pack_start (GTK_BOX (vbox), image, TRUE, TRUE, 0);
	gtk_widget_set_size_request (GTK_WIDGET (image), dim.screen.width, ( dim.screen.height / 100 ) * option.height);
	// ^ is it still needed??

	/* Bottom Separator */
	separator_bottom = gtk_hseparator_new ();
	gtk_box_pack_start (GTK_BOX (vbox), separator_bottom, FALSE, FALSE, 0);

	/* Timeout */
	g_timeout_add (1000 / option.fps, on_timeout, image);

	gtk_widget_show_all (mag_window);

	gtk_window_get_size (GTK_WINDOW (mag_window), &dim.win.width, &dim.win.height);
	move_magnifier ();

	// Reserve space
    vals[0] = 0;																//left
    vals[1] = 0;																//right
    vals[2] = (option.position == TOP) ? dim.win.height + option.offset : 0;	//top
    vals[3] = (option.position == BOTTOM) ? dim.win.height + option.offset : 0;	//bottom
    vals[8] = 0;																// top_start_x
    vals[9] = (option.position == TOP) ? dim.screen.height - 1 : 0;				// top_end_x
	vals[10] = 0;																// bottom_start_x
	vals[11] = (option.position == BOTTOM) ? dim.screen.height - 1 : 0;			// bottom_end_x

	gdk_property_change(
		mag_window->window,
		gdk_atom_intern ("_NET_WM_STRUT_PARTIAL", FALSE),
		gdk_atom_intern ("CARDINAL", FALSE),
		32,
		GDK_PROP_MODE_REPLACE,
		(guchar *)vals,
		12
	);

	gdk_property_change (
		mag_window->window,
		gdk_atom_intern ("_NET_WM_STRUT", FALSE),
		gdk_atom_intern ("CARDINAL", FALSE), 
		32,
		GDK_PROP_MODE_REPLACE,
		(guchar *)vals,
		4
	);

}

/* ------------------------------------------------------------------------- */

void cb_cbt_zoom (GtkComboBox *widget, gpointer data) {

	option.zoom = zoom_levels[gtk_combo_box_get_active (widget)];
	create_area_pixbuf ();

	zoom_levels_current = 0;
	while((zoom_levels_current < zoom_levels_size) && (zoom_levels[zoom_levels_current] != option.zoom))
		zoom_levels_current++;

}

/* ------------------------------------------------------------------------- */

void cb_cbt_position (GtkComboBox *widget, gpointer data) {

	option.position = gtk_combo_box_get_active (widget);
	mag_restart = TRUE;

}

/* ------------------------------------------------------------------------- */

void cb_sb_offset (GtkSpinButton *widget, gpointer data) {

	option.offset = gtk_spin_button_get_value (widget);
	mag_restart = TRUE;

}

/* ------------------------------------------------------------------------- */

void cb_sb_height (GtkSpinButton *widget, gpointer data) {

	option.height = gtk_spin_button_get_value (widget);
	create_area_pixbuf ();
	mag_restart = TRUE;

}

/* ------------------------------------------------------------------------- */

void cb_sb_fps (GtkSpinButton *widget, gpointer data) {

	option.fps = gtk_spin_button_get_value (widget);
	mag_restart = TRUE;

}

/* ------------------------------------------------------------------------- */

void cb_chk_show_magnifier (GtkToggleButton *widget, gpointer window) {

	gboolean show_mag_current;

	show_mag_current = option.show_magnifier;
	option.show_magnifier = gtk_toggle_button_get_active (widget);

	if (show_mag_current != option.show_magnifier) {

		if (option.show_magnifier == TRUE) {

			mag_restart = FALSE;
			show_magnifier();

		} else {

			gtk_widget_destroy (mag_window);

		}

	}

}

/* ------------------------------------------------------------------------- */

void cb_chk_hide_on_hover (GtkToggleButton *widget, gpointer window) {

	option.hide_on_hover = gtk_toggle_button_get_active (widget);

}

/* ------------------------------------------------------------------------- */

void cb_chk_show_cursor (GtkToggleButton *widget, gpointer data) {

	option.show_cursor = gtk_toggle_button_get_active (widget);
	create_area_pixbuf ();
	mag_restart = TRUE;

}

/* ------------------------------------------------------------------------- */

void cb_chk_interpol (GtkToggleButton *widget, gpointer data) {

	option.interpolation = gtk_toggle_button_get_active (widget);

}

/* ------------------------------------------------------------------------- */

void cb_btn_save (GtkWidget *widget, gpointer data) {

	save_options();
	gtk_widget_destroy (data);

}

/* ------------------------------------------------------------------------- */

GtkWidget *helper_mk_label (GtkWidget *target, gchar *label, gfloat xalign, gfloat yalign, gint x1, gint x2, gint y1, gint y2) {

	GtkWidget *widget;

	widget = gtk_label_new (label);
	gtk_misc_set_alignment (GTK_MISC (widget), xalign, yalign);
	gtk_table_attach_defaults (GTK_TABLE (target), widget, x1, x2, y1, y2);

	return widget;
}

/* ------------------------------------------------------------------------- */

void show_preferences (GtkWidget *widget, gpointer data) {

	GtkWidget *window, *vbox, *hbox, *table, *separator;
	GtkWidget *cbt_zoom, *cbt_position, *sb_offset, *sb_height, *sb_fps;
	GtkWidget *chk_show_magnifier, *chk_hide_on_hover, *chk_cursor, *chk_interpol;
	GtkWidget *btn_save, *btn_close;

	// Window
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW (window), _("Preferences"));
	gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
	gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
	gtk_window_set_icon_name (GTK_WINDOW (window), "gtk-preferences");

	// Vbox
	vbox = gtk_vbox_new (FALSE, 3);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 4);
	gtk_container_add (GTK_CONTAINER (window), vbox);

	// Table
	table = gtk_table_new (5, 2, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE (table), 8);
	gtk_table_set_row_spacings (GTK_TABLE (table), 4);
	gtk_container_set_border_width (GTK_CONTAINER (table), 4);
	gtk_container_add (GTK_CONTAINER (vbox), table);

	// Zoom
	helper_mk_label (table, _("Zoom:"), 0, 0.5, 0, 1, 0, 1);

	cbt_zoom = gtk_combo_box_new_text ();
	for (gint i = 0; i < zoom_levels_size; ++i) {
		gtk_combo_box_append_text (GTK_COMBO_BOX (cbt_zoom), g_strdup_printf ("%.1fx", zoom_levels[i]));
	}
	gtk_combo_box_set_active (GTK_COMBO_BOX (cbt_zoom), zoom_levels_current);
	g_signal_connect (G_OBJECT (cbt_zoom), "changed", G_CALLBACK (cb_cbt_zoom), NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), cbt_zoom, 1, 2, 0, 1);

	// Position
	helper_mk_label (table, _("Position:"), 0, 0.5, 0, 1, 1, 2);

	cbt_position = gtk_combo_box_new_text ();
	gtk_combo_box_append_text(GTK_COMBO_BOX (cbt_position), _("Top"));
	gtk_combo_box_append_text(GTK_COMBO_BOX (cbt_position), _("Bottom"));
	gtk_combo_box_set_active (GTK_COMBO_BOX (cbt_position), option.position);
	g_signal_connect (G_OBJECT (cbt_position), "changed", G_CALLBACK (cb_cbt_position), NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), cbt_position, 1, 2, 1, 2);

	// Offset
	helper_mk_label (table, _("Offset (px):"), 0, 0.5, 0, 1, 2, 3);

	sb_offset = gtk_spin_button_new_with_range (OFFSET_MIN, OFFSET_MAX, OFFSET_STEP);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (sb_offset), option.offset);
	g_signal_connect (G_OBJECT (sb_offset), "value-changed", G_CALLBACK (cb_sb_offset), NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), sb_offset, 1, 2, 2, 3);

	// Height
	helper_mk_label (table, _("Height (%):"), 0, 0.5, 0, 1, 3, 4);

	sb_height = gtk_spin_button_new_with_range (HEIGHT_MIN, HEIGHT_MAX, HEIGHT_STEP);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (sb_height), option.height);
	g_signal_connect (G_OBJECT (sb_height), "value-changed", G_CALLBACK (cb_sb_height), NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), sb_height, 1, 2, 3, 4);

	// FPS
	helper_mk_label (table, _("FPS:"), 0, 0.5, 0, 1, 4, 5);

	sb_fps = gtk_spin_button_new_with_range (FPS_MIN, FPS_MAX, FPS_STEP);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (sb_fps), option.fps);
	g_signal_connect (G_OBJECT (sb_fps), "value-changed", G_CALLBACK (cb_sb_fps), NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), sb_fps, 1, 2, 4, 5);

	// Show Magnifier
	chk_show_magnifier = gtk_check_button_new_with_label (_("Show magnifier"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chk_show_magnifier), option.show_magnifier);
	g_signal_connect (G_OBJECT (chk_show_magnifier), "clicked", G_CALLBACK (cb_chk_show_magnifier), NULL);
	gtk_box_pack_start (GTK_BOX (vbox), chk_show_magnifier, TRUE, TRUE, 0);

	// Hide On Hover
	chk_hide_on_hover = gtk_check_button_new_with_label (_("Hide on mouse hover"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chk_hide_on_hover), option.hide_on_hover);
	g_signal_connect (G_OBJECT (chk_hide_on_hover), "clicked", G_CALLBACK (cb_chk_hide_on_hover), NULL);
	gtk_box_pack_start (GTK_BOX (vbox), chk_hide_on_hover, TRUE, TRUE, 0);

	// Show Cursor
	chk_cursor = gtk_check_button_new_with_label (_("Show mouse pointer"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chk_cursor), option.show_cursor);
	g_signal_connect (G_OBJECT (chk_cursor), "clicked", G_CALLBACK (cb_chk_show_cursor), NULL);
	gtk_box_pack_start (GTK_BOX (vbox), chk_cursor, TRUE, TRUE, 0);

	// Interpolation
	chk_interpol = gtk_check_button_new_with_label (_("Interpolation (CPU intensive!)"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (chk_interpol), option.interpolation);
	g_signal_connect (G_OBJECT (chk_interpol), "clicked", G_CALLBACK (cb_chk_interpol), NULL);
	gtk_box_pack_start (GTK_BOX (vbox), chk_interpol, TRUE, TRUE, 0);

	// Separator
	separator = gtk_hseparator_new ();
	gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, FALSE, 0);

	hbox = gtk_hbox_new (FALSE, 4);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	// Save Button
	btn_save = gtk_button_new_from_stock ("gtk-save");
	g_signal_connect (G_OBJECT (btn_save), "clicked", G_CALLBACK (cb_btn_save), window);
	gtk_box_pack_start (GTK_BOX (hbox), btn_save, TRUE, TRUE, 0);

	// Close Button
	btn_close = gtk_button_new_from_stock ("gtk-close");
	g_signal_connect_swapped (G_OBJECT (btn_close), "clicked", G_CALLBACK (gtk_widget_destroy), window);
	gtk_box_pack_start (GTK_BOX (hbox), btn_close, TRUE, TRUE, 0);

	gtk_widget_show_all (window);

}

/* ------------------------------------------------------------------------- */

void show_help (GtkWidget *widget, gpointer data) {
	
	GtkWidget *window, *vbox, *btn_close, *view;
	GtkTextBuffer *buffer;

	const gchar *help_msg = _( \
"Tray icon mouse actions:\n\n\
Left button:\t\ttoggle magnifier window\n\
Middle button:\ttoggle mouse pointer\n\
Right button:\t\tshow menu\n\
Wheel:\t\t\tadjust magnification\n\
Shift+Wheel:\t\tadjust magnifier height\n"
	);

	// Window
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (window), _("Help"));
	gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
	gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER);
	gtk_window_set_icon_name (GTK_WINDOW (window), "gtk-help");

	// Vbox
	vbox = gtk_vbox_new (FALSE, 3);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 4);
	gtk_container_add (GTK_CONTAINER (window), vbox);

	// TextView
	view = gtk_text_view_new ();
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
	gtk_text_buffer_set_text (buffer, help_msg, -1);
	gtk_text_view_set_editable (GTK_TEXT_VIEW (view), FALSE);
	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (view), FALSE);
	gtk_text_view_set_left_margin (GTK_TEXT_VIEW (view), 8);
	gtk_text_view_set_right_margin (GTK_TEXT_VIEW (view), 8);
	gtk_box_pack_start (GTK_BOX (vbox), view, TRUE, TRUE, 3);

	// Close Button
	btn_close = gtk_button_new_from_stock ("gtk-close");
	g_signal_connect_swapped (G_OBJECT (btn_close), "clicked", G_CALLBACK (gtk_widget_destroy), window);
	gtk_box_pack_start (GTK_BOX (vbox), btn_close, TRUE, TRUE, 0);

	gtk_widget_show_all (window);

}
	
/* ------------------------------------------------------------------------- */

void show_about (GtkWidget *widget, gpointer data) {

	const gchar *appname	= APPNAME;
	const gchar *comments	= "A Screen Magnifier Dock";
	const gchar *copyright	= "Copyright © JakeSFR 2019";
	const gchar *authors[]	= { "JakeSFR", '\0' };
	const gchar *artists[]	= { \
		_("Program icon assembled by JakeSFR"),
		_("using vovchik's Pikona and its built-in icons."),
		"http://www.murga-linux.com/puppy/viewtopic.php?t=49129",
		'\0'
	};
	const gchar *license 	= _( \
"This program is free software: you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation, either version 3 of the License, or \
(at your option) any later version. \
\n\n\
This program is distributed in the hope that it will be useful, \
but WITHOUT ANY WARRANTY; without even the implied warranty of \
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \
GNU General Public License for more details. \
\n\n\
You should have received a copy of the GNU General Public License \
along with this program.  If not, see <http://www.gnu.org/licenses/>."
	);

	gtk_window_set_default_icon_name ("gtk-about");

	gtk_show_about_dialog (
		NULL,
		"program-name",	appname,
		"authors",		authors,
		"artists",		artists,
		"comments",		comments,
		"copyright",	copyright,
		"license",		license,
		"wrap license",	TRUE,
		"title",		_("About"),
		NULL
	);

}

/* ------------------------------------------------------------------------- */

void cb_toggle_magnifier (GtkStatusIcon *status_icon, GdkEvent *event, gpointer data) {

	if (option.show_magnifier == FALSE) {

		mag_restart = FALSE;
		show_magnifier ();
		option.show_magnifier = TRUE;

	} else {

		gtk_widget_destroy (mag_window);
		option.show_magnifier = FALSE;

	}

}

/* ------------------------------------------------------------------------- */

gboolean cb_button (GtkStatusIcon *status_icon, GdkEventButton *event, gpointer data) {

	if (option.show_magnifier != TRUE) return FALSE;

	if (event->button == 2) {

		option.show_cursor ^= 1;
		create_area_pixbuf ();
		mag_restart = TRUE;

		return TRUE;

	} else {

		return FALSE;

	}

}

/* ------------------------------------------------------------------------- */

void cb_scroll (GtkStatusIcon *status_icon, GdkEventScroll *event, gpointer data) {

	gint old_height;

	if (option.show_magnifier != TRUE) return;

	// Zoom
	if ((event->state & GDK_SHIFT_MASK) == 0) {
	
		if (event->direction == GDK_SCROLL_UP) {

			if (option.zoom < zoom_levels[zoom_levels_size-1]) {
				option.zoom = zoom_levels[++zoom_levels_current];
			}

		} else if (event->direction == GDK_SCROLL_DOWN) {

			if (option.zoom > zoom_levels[0]) {
				option.zoom = zoom_levels[--zoom_levels_current];
			}

		} else {

			return;

		}

		create_area_pixbuf ();	

	// Height
	} else {

		old_height = option.height;

		if (event->direction == GDK_SCROLL_UP) {

			option.height = (option.height < HEIGHT_MAX) ? option.height + HEIGHT_STEP : HEIGHT_MAX;

		} else if (event->direction == GDK_SCROLL_DOWN) {

			option.height = (option.height > HEIGHT_MIN) ? option.height - HEIGHT_STEP : HEIGHT_MIN;

		}

		if (option.height != old_height) {

			create_area_pixbuf ();
			mag_restart = TRUE;

		}

	}

}

/* ------------------------------------------------------------------------- */

void cb_show_menu (GtkStatusIcon *status_icon, guint button, guint time, gpointer menu) {

	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, button, time);

}

/* ------------------------------------------------------------------------- */

GtkWidget *helper_mk_menuitem_stock (gchar *stock, gchar *label, GtkWidget *target, gpointer callback) {

	GtkWidget *widget;

	widget = gtk_image_menu_item_new_from_stock (stock, NULL);
	if (label) gtk_menu_item_set_label (GTK_MENU_ITEM (widget), label);
	gtk_menu_shell_append (GTK_MENU_SHELL (target), widget);
	g_signal_connect (G_OBJECT (widget), "activate", G_CALLBACK (callback), NULL);

	return widget;
}

/* ------------------------------------------------------------------------- */

void show_tray_icon (void) {

	GtkStatusIcon	*status_icon;
	GtkWidget		*menu, *separator, *title;

	menu = gtk_menu_new ();

	title = gtk_image_menu_item_new_with_label (APPNAME);
	gtk_widget_set_sensitive (GTK_WIDGET (title), FALSE);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), title);

	separator = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), separator);

    helper_mk_menuitem_stock ("gtk-preferences",	NULL, menu, show_preferences);
    helper_mk_menuitem_stock ("gtk-help",			NULL, menu, show_help);
    helper_mk_menuitem_stock ("gtk-about",			NULL, menu, show_about);
    helper_mk_menuitem_stock ("gtk-quit",			NULL, menu, gtk_main_quit);


#if EMBED_ICON == TRUE
	#include "magdock.xpm"
	GdkPixbuf *icon = gdk_pixbuf_new_from_xpm_data (magdock_xpm);
	status_icon = gtk_status_icon_new_from_pixbuf (icon);
	g_object_unref (icon);
#else
	status_icon = gtk_status_icon_new_from_icon_name ("magdock");
#endif
	gtk_status_icon_set_tooltip_text (status_icon, APPNAME);
	gtk_status_icon_set_visible (status_icon, FALSE);
	g_signal_connect (G_OBJECT(status_icon), "activate", G_CALLBACK (cb_toggle_magnifier), NULL);
	g_signal_connect (G_OBJECT(status_icon), "popup-menu", G_CALLBACK (cb_show_menu), menu);
	g_signal_connect (G_OBJECT(status_icon), "scroll_event", G_CALLBACK (cb_scroll), NULL);
	g_signal_connect (G_OBJECT(status_icon), "button-press-event", G_CALLBACK (cb_button), NULL);
	g_signal_connect (menu, "destroy", G_CALLBACK (gtk_main_quit), NULL);

	gtk_widget_show_all (menu);

	// Show the icon _after_ showing the widget itself, otherwise the icon
	// gets broken under JWM built in/for Slacko
	gtk_status_icon_set_visible (status_icon, TRUE);

	if (option.show_magnifier == TRUE) {
		show_magnifier ();
	}

	gtk_main ();

}

/* ------------------------------------------------------------------------- */

gint main (gint argc, gchar **argv) {

	// Gettext
	const gchar *domain = "magdock";

	setlocale (LC_ALL, "");
	bindtextdomain (domain, "/usr/share/locale");
	textdomain (domain);
	bind_textdomain_codeset (domain, "UTF-8");

	// Load Options
	load_options ();

	// Initialize GTK
	gtk_init (&argc, &argv);

	// Create Main Pixbuf
	create_area_pixbuf ();

	// Show Tray Icon
	show_tray_icon ();

	// Unref main pixbuf
	g_object_unref (dim.area.pixbuf);

	return 0;
}

/*****************************************************************************/
